package org.tapomay;

import java.beans.PropertyVetoException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import com.mchange.v2.c3p0.ComboPooledDataSource;

/**
 * Acquires a lock on the master before executing DML.
 * This is because monetdb uses fully serialized transactions 
 * that throw errors on concurrent commits.
 * @author tapomay
 *
 */
public class JdbcMonetdbTestClient
{
    private static final String DRIVER_NAME = 
        "nl.cwi.monetdb.jdbc.MonetDriver";
    
    private DataSource dataSource;
    
    private enum QueryErrorType
    {
        CONN_ERROR,
        SQL_ERROR,
        NO_ERROR
    }
    
    public JdbcMonetdbTestClient(String url, String username, String password)
    {
        initConnectionPool(url, username, password);
    }

    private void initConnectionPool(String url, String username, String password)
    {
        ComboPooledDataSource cpds = new ComboPooledDataSource();
        try
        {
            cpds.setDriverClass(DRIVER_NAME);
        }
        catch (PropertyVetoException e)
        {
            e.printStackTrace();
        }
        
        
        cpds.setJdbcUrl(url);
        cpds.setUser(username);
        cpds.setPassword(password);

        cpds.setInitialPoolSize(1);
        cpds.setMinPoolSize(10);
        cpds.setAcquireIncrement(5);
        cpds.setMaxPoolSize(20);
        cpds.setIdleConnectionTestPeriod(60);
        cpds.setMaxStatements(0);
        cpds.setCheckoutTimeout(1000);
        cpds.setPreferredTestQuery("select 1;");// To prevent hang!
//        try
//        {
//            cpds.setConnectionTesterClassName("com.sokrati.monetawrapper.c3p0.MonetaC3p0ConnectionTester");
//        }
//        catch (PropertyVetoException e)
//        {
//            throw new RuntimeException("Failed to set c3p0 conenction tester classname.", e);
//        }
        dataSource = cpds;
    }
    
    public synchronized Connection startNewTransaction() throws SQLException
    {
        Connection connection = null;
        connection = dataSource.getConnection();
        connection.setAutoCommit(false);
        return connection;
    }

    public int executeUpdate(String updateQuery)
    {
        Connection connection = null;
        try
        {
            connection = dataSource.getConnection();
        }
        catch (SQLException e)
        {
            System.out.println(e);
        }

        return executeUpdate(updateQuery, connection, true);
        
    }
    
    public int executeUpdate(String updateQuery, Connection connection, boolean closeConnection)
    {
        int rowCount = -1;
        try
        {
            if (null != connection)
            {
                Statement statement = connection.createStatement();
                rowCount = statement.executeUpdate(updateQuery);
            }
        }
        catch (SQLException e)
        {
            QueryErrorType errorType = parseSqlException(e.getMessage());
            switch (errorType)
            {
                case CONN_ERROR:
                    break;
                case SQL_ERROR:
                    System.out.println(QueryErrorType.SQL_ERROR);
                    break;
                case NO_ERROR:
                    break;
                default:
            }
        }
        finally
        {
            try
            {
                if (closeConnection && null != connection)
                {
                    connection.close();
                }
            }
            catch (SQLException e)
            {
            }
        }
        return rowCount;
    }

    private QueryErrorType parseSqlException(String errMsg)
    {
        QueryErrorType errorType = QueryErrorType.SQL_ERROR;
        
        if (errMsg.contains("Connection to server lost!"))
        {
            errorType = QueryErrorType.CONN_ERROR;
        }
        else if (errMsg.contains("already in use") ||
                 errMsg.contains("already exists"))
        {
            errorType = QueryErrorType.NO_ERROR;
        }
        
        return errorType;
    }
    
    public List<UniqueKey> getUniqueKeys(String tableName)
        throws SQLException
    {
        List<UniqueKey> list = new ArrayList<UniqueKey>();
        Map<Integer, UniqueKey> uniqueKeyMap = new HashMap<Integer, UniqueKey>();
        
        Connection connection = null;
        StringBuilder query = new StringBuilder();
        query.append("select k.id as key, o.name as col from sys.keys k, sys.objects o where k.id");
        query.append(" in (select id from sys.keys where table_id");
        query.append(" = (select id from sys.tables where name='");
        query.append(tableName);
        query.append("') and name like '%unique%') and o.id = k.id");
        try
        {
            connection = dataSource.getConnection();
            Statement statement = connection.createStatement();
            ResultSet result = statement.executeQuery(query.toString());
            while (result.next())
            {
                int key = result.getInt("key");
                UniqueKey uniqueKey = uniqueKeyMap.get(key);
                if(null == uniqueKey)
                {
                    uniqueKey = new UniqueKey(key);
                    uniqueKeyMap.put(key, uniqueKey);
                }
                String col = result.getString("col");
                uniqueKey.add(col);
            }
            Collection<UniqueKey> values = uniqueKeyMap.values();
            list.addAll(values);
        }
        catch (SQLException e)
        {

        }
        finally
        {
            try
            {
                if (null != connection)
                {
                    connection.close();
                }
            }
            catch (SQLException e)
            {
            }
        }
        return list;
    }

    public static void printResultSet(ResultSet rs) throws SQLException
    {
        System.out.print("RS: ");
        ResultSetMetaData rsmd = rs.getMetaData();
        
        int numberOfColumns = rsmd.getColumnCount();
        
        while (rs.next()) {
          for (int i = 1; i <= numberOfColumns; i++) {
            if (i > 1) System.out.print(",  ");
            String columnValue = rs.getString(i);
            System.out.print(i + ":" +columnValue);
          }
          System.out.println("");  
        }
    
    }

}
