package org.tapomay;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

public class Multiplexer
{
    private ExecutorService executorService;
    private static Multiplexer _instance;

    private Multiplexer()
    {
        executorService = Executors.newCachedThreadPool();
    }

    public synchronized static Multiplexer getInstance()
    {
        if (null == _instance)
        {
            _instance = new Multiplexer();
        }
        return _instance;
    }

    /**
     * Expects set-list one:one correspondence between a master and its
     * connection.
     * 
     * @param updateQuery
     * @param liveMasterSet
     * @param connectionList
     * @return
     * @throws QueryExecutionException
     */
    public Integer multiplexUpdate(String updateQuery,
                                       Set<JdbcMonetdbTestClient> liveMasterSet,
                                       List<Connection> connectionList)
    {

        Map<JdbcMonetdbTestClient, Future<Integer>> futureMap =
                new HashMap<JdbcMonetdbTestClient, Future<Integer>>();

        int parties = liveMasterSet.size() + 1;
        // Adding 1 to account for Multiplexer calling barrier.await();
        CyclicBarrier barrier = new CyclicBarrier(parties);

        try
        {
            int i = 0;
            for (JdbcMonetdbTestClient master : liveMasterSet)
            {
                Connection connection = connectionList.get(i++);
                MonetdbClientCallable callable =
                    new MonetdbClientCallable(master, updateQuery,
                        barrier, connection);

                Future<Integer> future =
                    executorService.submit(callable);
                futureMap.put(master, future);
            }
        }
        finally
        {
            try
            {
                barrier.await();
            }
            catch (InterruptedException e)
            {
            }
            catch (BrokenBarrierException e)
            {
            }
        }
        
        Integer ret = 0;

        for (JdbcMonetdbTestClient master : liveMasterSet)
        {
            Future<Integer> future = futureMap.get(master);
            if (null == future)
            {
                continue;
            }
            try
            {
                Integer queryResult = future.get();
                ret += queryResult;
            }
            catch (InterruptedException e)
            {
            }
            catch (ExecutionException e)
            {
            }
        }

        return ret;
    }

    /**
     * Returns number of connections committed successfully.
     * @param connectionList
     * @return
     * @throws SQLException
     */
//    public int commit(List<Connection> connectionList) throws SQLException
//    {
//        int ret = 0;
//        for (Connection c : connectionList)
//        {
//            c.commit();
//            ret ++;
//        }
//        return ret;
//    }

    /**
     * Rollback connections ignoring any failures.
     * @param connectionList
     */
    public static void rollback(List<Connection> connectionList)
    {
        for (Connection c : connectionList)
        {
            try
            {
                c.rollback();
            }
            catch (SQLException e)
            {
                //NOP
            }
        }
    }

    /**
     * Close connections ignoring any failures.
     * @param connectionList
     */
    public static void close(List<Connection> connectionList)
    {
        for (Connection c : connectionList)
        {
            try
            {
                c.close();
            }
            catch (SQLException e)
            {
                //NOP
            }
        }
    }
    
}
